
/* rdm_in Library (RDM responder lib, C version)
 ****************************************************************************
 * Author:      Hendrik Hlscher                         					*
 * modified by:    															*
 * Revision:    1.3                              							*
 * Date:        11.05.21                  									*
 * Target:      Transceiver Rev.3.01 (www.hoelscher-hi.de)        			*
 * Compiled using Atmel Studio 7                								*
 ****************************************************************************
 * Fuses:    OSC:  8.0 Mhz xtal	          								*
 *           WDT:  on                       								*
 ****************************************************************************
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version2 of
 the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 General Public License for more details.

 If you have no copy of the GNU General Public License, write to the
 Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 For other license models, please contact the author.	              		*
 ****************************************************************************
 */

#include "globals.h"
#include "lib_rdm_in.h"
#include <string.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>

// ********************* local definitions *********************

enum {IDLE, BRK, SB_DMX, STARTADR_DMX, SB_RDM, SSC_RDM, LEN_RDM, CHECKH_RDM, CHECKL_RDM, RESPOND};			//Rx states

#define UID_CS (0xFF *6 +MAN_H_ID +MAN_L_ID +0xFF +TYPE_ID +0x68)			//prepare check sum @design time

typedef union {
   uint16_t u16;
   struct {
          uint8_t u8l;
          uint8_t u8h;
          };
} uint16or8;


volatile uint8_t 	 RxState= IDLE;
		 uint8_t	 DevID[8];							//MAC Address of Device
		 uint16or8	 RxCount_16;						//start address and check sum counting
		 uint8_t     RxCount_8;							//byte counting
		 uint8_t	 gTxCh;
		 uint8_t	 TxLen;
		 uint8_t	 RdmFlags;

const uint8_t DiscRespMsg[] PROGMEM = {					//discovery response msg preamble
0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xAA
};

const uint8_t InfoMsg[] PROGMEM = {					   	//device info msg
1, 0,													//RDM protocol version
TYPE_ID, 0, 											//device model ID
(FIXT_DIM_LED>>8),										//fixture type
(FIXT_DIM_LED &0xFF),
0, 0, 2, 0,										//FIRMWARE VERSION
0, DMX_CHANNELS,										//DMX footprint
1, 2,													//current (dummy) DMX personality, all DMX personalities				
0, 1,													//dummy start address
0, 0,													//no of sub devices
1														//no of sensors
};


const uint8_t ManufactMsg[] PROGMEM = {					//Manufacturer Label
'H','e','n','n','e'
};


const uint8_t StatusOkMsg[] PROGMEM = {
0, 0,														//sub device ID
STATUS_NONE, 0,												//device state
STS_READY,													//error ID
0x00, 0x00,  												//Parameter 1
0x00, 0x00													//Parameter 2
};

const uint8_t StatusHotMsg[] PROGMEM = {
0, 0,														//sub device ID
STATUS_WARNING, 0,											//device state
STS_OVERTEMP,												//error ID
0x00, 0x01,  												//Parameter 1 (Sensor ID)
0x00, 75													//Parameter 2 (Temperature)
};


const uint8_t ParamMsg[] PROGMEM = {					//supported parameters msg
(PERSONALITY>>8),          (PERSONALITY &0xFF),
(DMX_START_ADDRESS>>8),    (DMX_START_ADDRESS &0xFF),
(DEVICE_LABEL>>8),         (DEVICE_LABEL &0xFF),
(MANUFACT_LABEL>>8),       (MANUFACT_LABEL &0xFF),
(STATUS_MESSAGES>>8),      (STATUS_MESSAGES &0xFF),
(MIN_LIMIT>>8),    		   (MIN_LIMIT &0xFF),
(MAX_LIMIT>>8),    		   (MAX_LIMIT &0xFF),
(ERR_SCENE>>8),    		   (ERR_SCENE &0xFF)
};



const uint8_t MinDescr[] PROGMEM = {					//description msg of min limit
DIMM_CH,												//size
DS_UNSIGNED_BYTE,										//data type
CC_GET_SET,												//command class
0x7F,													//type
0,														//unit
0,														//prefix
0,0,0,0,												//minimum [8]..[11]
0,0,0,64,												//maximum [12]..[15]
0,0,0,0,												//default [16]..[19]
'P','r','e','h','e','a','t'								//description
};

const uint8_t MaxDescr[] PROGMEM = {					//description msg of max limit
DIMM_CH,												//size
DS_UNSIGNED_BYTE,										//data type
CC_GET_SET,												//command class
0x7F,													//type
0,														//unit
0,														//prefix
0,0,0,191,												//minimum [8]..[11]
0,0,0,0xFF,												//maximum [12]..[15]
0,0,0,0xFF,												//default [16]..[19]
'T','o','p','s','e','t'									//description
};

const uint8_t ErrDescr[] PROGMEM = {					//description msg of max limit
DIMM_CH,												//size
DS_UNSIGNED_BYTE,										//data type
CC_GET_SET,												//command class
0x7F,													//type
0,														//unit
0,														//prefix
0,0,0,0,												//minimum [8]..[11]
0,0,0,0xFF,												//maximum [12]..[15]
0,0,0,0x80,												//default [16]..[19]
'E','r','r','S','c','e','n','e'							//description
};



// *************** DMX Reception Initialisation ****************
void start_DMX_Rx(void)
{
PORTD &= ~(1<<PD2);									//enable receiver
UBRRH  = 0;
UBRRL  = 1;											//250kbaud, 8N2
UCSRC  = (1<<URSEL)|(3<<UCSZ0)|(1<<USBS);
UCSRB  = (1<<RXEN)|(1<<RXCIE);
RxState = IDLE;
RdmIdentifyCnt= 0;									//no RDM identify

uint8_t i= UDR;										//clear flags
for (i=0; i<DMX_CHANNELS; i++)						//reset field
	{
	DmxField[i]= 0;
	}
}



// **** helper functions ****
void delay(uint8_t cnt)
{
while (cnt != 0)
	{
	if (gFlags &(1<<TICK))							//called with ~100Hz
		{
		gFlags &= ~(1<<TICK);
		wdt_reset();
		cnt--;
		}
	}
}


static void getDevId(void)								//get rdm device uid
{
DevID[0]= MAN_H_ID;									//manufacturer ID (registered by ESTA)
DevID[1]= MAN_L_ID;
DevID[2]= 0xFF;										//(don't change for Henne's Sites)
DevID[3]= TYPE_ID;									//ID of device type 
DevID[4]= 0x68;										//can be changed...
DevID[5]= 0xFF -PINC;								//use DIPs as part of RDM-UID

uint16or8 buf;
buf.u16= UID_CS  +(uint16_t)DevID[5];
DevID[6]= buf.u8h;									//add check sum
DevID[7]= buf.u8l;
}


static inline uint16_t swapInt(uint16_t val)		//change byte order
{
uint16_t res  = (val<<8);
		 res |= (val>>8);
return (res);
}

static uint8_t forUs(void)
{
struct RDM_Packet *rdm;
rdm = (struct RDM_Packet *)&RdmField;
uint8_t i, buf, res= 0, resB= 0;
for (i=0; i<6; i++)
	{
	buf= rdm->DestID[i];
	if (buf != DevID[i]) res= 1;
	if (buf != 0xFF)     resB= 1;
	}
if (resB == 0) res= resB;
return (res);
}


static uint8_t isLower(void)
{
struct RDM_Packet *rdm;
rdm = (struct RDM_Packet *)&RdmField;
uint8_t i ,buf;
for (i=0; i<6; i++)
	{
	buf= rdm->Data[i]; 
	if      (buf < DevID[i]) return (0);
	else if (buf > DevID[i]) return (1);
	}
return (0);
}



static uint8_t isHigher(void)
{
struct RDM_Packet *rdm;
rdm = (struct RDM_Packet *)&RdmField;
uint8_t i ,buf;
for (i=0; i<6; i++)
	{
	buf= rdm->Data[i +6]; 
	if      (buf > DevID[i]) return (0);
	else if (buf < DevID[i]) return (1);
	}
return (0);
}


// ****************** RDM data handling *******************
void respondMsg(void)
{
struct RDM_Packet *rdm;									//set addresses
rdm = (struct RDM_Packet *)&RdmField;

PORTD |= (1<<PD2);									//enable transmission
uint8_t i;
for (i=0; i<6; i++)
	{
	rdm->DestID[i]= rdm->SrcID[i];					//controller is target
	rdm->SrcID[i]= DevID[i];
	}
rdm->ASC=      0xCC;
rdm->SSC=      0x01;
TxLen=   	   rdm->PDLen +24;
rdm->Length=   TxLen;
TxLen++;											//transmit complete checksum
rdm->MsgCount= 0;									//currently no queued msgs supported
rdm->Cmd=      rdm->Cmd +1;							//it is a response... 

for (i=0; i<18; i++)								//wait to allow port turn around
   {
	_delay_us(10);
   }

UBRRL  = 10;										//45.5kbaud
gTxCh  = 0;											//reset field pointer
UCSRB  = (1<<TXEN)|(1<<TXCIE);
UDR    = 0;											//send break
RxState= RESPOND;

uint16or8 cs;										//build check sum
cs.u16= 0;
for (i=0; i<rdm->Length; i++)
	{
	cs.u16+= (uint16_t)RdmField[i];
	}
RdmField[rdm->Length]= cs.u8h;
RdmField[rdm->Length +1]= cs.u8l;
}


// *** generate "discover unique branch" response packet
void transmit_byte(uint8_t val)
{
_delay_us(5);
UDR  =val;
UCSRA= (1<<TXC);
loop_until_bit_is_set(UCSRA, TXC);
}

static void respondDisc(void)
{
if (!(RdmFlags &(1<<MUTE_RDM)))
	{
	uint8_t i;
	PORTD  |= (1<<2);										//enable transmission
	for (i=0; i<18; i++)									//wait to allow port turn around
	   {
		_delay_us(10);
	   }
	
	UBRRH	= 0;
	UBRRL   = 1;											//250kbaud
	UCSRB   = (1<<TXEN);
	for(i=0;i<sizeof(DiscRespMsg);i++) 
		{
    	transmit_byte(pgm_read_byte(&DiscRespMsg[i]));		//transmit preamble					
		}
	wdt_reset();
	uint8_t buf;
	for(i=0;i<8;i++) 
		{
		buf  = DevID[i];
    	transmit_byte(buf |0xAA);							//transmit encoded response
    	transmit_byte(buf |0x55);
		}
	wdt_reset();
	RxState= IDLE;    				   						//wait for break
	PORTD &= ~(1<<2);										//enable reception
	UCSRB  = (1<<RXEN)|(1<<RXCIE);
	}
}



void check_rdm(void)
{
uint16or8 tmp;
if (PINE &(1<<PE1))
	{
	//DMX mode
	tmp.u8l= 0xFF -PINC;									//invert DIP state
	if (PINE &(1<<PE2)) tmp.u8h= 0;
	else			    tmp.u8h= 1;
	if (tmp.u16 == 0)   tmp.u8l= 1;
	dur.DmxAddress= tmp.u16;
	}
else if (RdmFlags &(1<<EVAL_RDM))
	{
	getDevId();
	if (forUs() == 0) 										//is packet for us?
		{
		struct RDM_Packet *rdm;
		rdm = (struct RDM_Packet *)&RdmField;
		rdm->PortID= RESPONSE_TYPE_ACK;
		switch(swapInt(rdm->PID))
			{
			case DISC_UNIQUE_BRANCH:
				if((isLower() == 0) && (isHigher() == 0))	//is UID in disc branch?
					{
					respondDisc();							//answer discovery msg
					}
				break;				

			case DISC_MUTE:									//we shall shut up
				RdmFlags |= (1<<MUTE_RDM);
				rdm->Data[0]= 0;
				rdm->Data[1]= 0;
				rdm->PDLen=  2;
				respondMsg();								//ACK
				break;

			case DISC_UN_MUTE:								//we shall respond again
				RdmFlags &= ~(1<<MUTE_RDM);
				rdm->Data[0]= 0;
				rdm->Data[1]= 0;
				rdm->PDLen=  2;
				respondMsg();								//ACK
				break;

			case IDENTIFY:									//we shall respond again
				if (rdm->Cmd == SET_CMD)
					{
				    if (rdm->Data[0] == 1) RdmIdentifyCnt= 200;		//start identification (use a timeout to kill identification if controller hangs)
					else				   RdmIdentifyCnt= 0;		//stop  identification
					rdm->PDLen= 0;							//respond with no data
					}
				else
					{
					if (RdmIdentifyCnt == 0) rdm->Data[0]= 0;		//ACK with identify disabled
					else					 rdm->Data[0]= 1;		//ACK with identify enabled
				    rdm->PDLen= 1;
					}
				respondMsg();
				break;

			case DMX_START_ADDRESS:							//start address...
				if (rdm->Cmd == SET_CMD)
					{
					tmp.u8h= rdm->Data[0];
					tmp.u8l= rdm->Data[1];
					if      (tmp.u16 > 512) tmp.u16= dur.DmxAddress;	//limit to valid range (for safety)
					else if (tmp.u16 == 0)  tmp.u16= dur.DmxAddress;
					dur.DmxAddress= tmp.u16;				//change address
					}
				else
					{
					tmp.u16= dur.DmxAddress;
					rdm->Data[0]= tmp.u8h; 					//return current address
					rdm->Data[1]= tmp.u8l;
					}
				rdm->PDLen=  2;
				respondMsg();
				break;

			case PERSONALITY:								// personality (no of channels to be switched)
				if (rdm->Cmd == SET_CMD)
					{
					rdm->Data[0] -=1;
					if (rdm->Data[0] < 8)
						{
					    dur.Personality= rdm->Data[0];
						}
					}
				else
					{
					rdm->Data[0]= dur.Personality +1; 		//return current personality
					}
				rdm->Data[1]= 2;							//max 2 personalities
				rdm->PDLen=  2;
				respondMsg();
				break;


			case DEVICE_INFO:
				memcpy_P(&(rdm->Data),InfoMsg, sizeof(InfoMsg));
				tmp.u16= dur.DmxAddress;
				rdm->Data[12]= dur.Personality +1;			//return current personality
				rdm->Data[14]= tmp.u8h; 					//return current start address (profile missing)
				rdm->Data[15]= tmp.u8l;
				rdm->PDLen= sizeof(InfoMsg);
				respondMsg();
				break;

			case DEVICE_LABEL:
				if (rdm->Cmd == SET_CMD)					//set new device label
					{
					if (rdm->PDLen > 19) rdm->PDLen= 19;					//limit to prevent ee overflow
					dur.DevLabelLen= rdm->PDLen;
					memcpy(&dur.DevLabel[0], &rdm->Data[0], rdm->PDLen);
					}
				else
					{
					rdm->PDLen= dur.DevLabelLen;
					if (rdm->PDLen > 19) rdm->PDLen= 19;					//limit to prevent ee overflow
					memcpy(&rdm->Data[0], &dur.DevLabel[0], rdm->PDLen); //response with device label
					}
				respondMsg();
				break;

			case MANUFACT_LABEL:
				memcpy_P(&(rdm->Data),ManufactMsg, sizeof(ManufactMsg));
				rdm->PDLen= sizeof(ManufactMsg);
				respondMsg();
				break;
				
			case SOFTWARE_VERSION_LABEL:								//required by standard...
			    rdm->PDLen= 0;
			    respondMsg();
			    break;

			case STATUS_MESSAGES:
				if (ErrState == 2) memcpy_P(&(rdm->Data),StatusHotMsg, sizeof(StatusHotMsg));
				else			   memcpy_P(&(rdm->Data),StatusOkMsg, sizeof(StatusOkMsg));
				rdm->PDLen= sizeof(StatusOkMsg);
				respondMsg();
				break;

			case SUPPORTED_PARAMETERS:
				memcpy_P(&(rdm->Data),ParamMsg, sizeof(ParamMsg));
				rdm->PDLen= sizeof(ParamMsg);
				respondMsg();
				break;

			case PARAMETER_DESCRIPTION:
				if (rdm->Data[0] == (uint8_t)(MIN_LIMIT>>8))
					{
					if (rdm->Data[1] == (uint8_t)MIN_LIMIT)
						{
						memcpy_P(&(rdm->Data[2]),MinDescr, sizeof(MinDescr));
						rdm->PDLen= sizeof(MinDescr) +2;
						}
					else if (rdm->Data[1] == (uint8_t)MAX_LIMIT)
						{
						memcpy_P(&(rdm->Data[2]),MaxDescr, sizeof(MaxDescr));
						rdm->PDLen= sizeof(MaxDescr) +2;
						}
					else if (rdm->Data[1] == (uint8_t)ERR_SCENE)
						{
						memcpy_P(&(rdm->Data[2]),ErrDescr, sizeof(ErrDescr));
						rdm->PDLen= sizeof(ErrDescr) +2;
						}
					respondMsg();
					}
				break;
			
			case MIN_LIMIT:
				if (rdm->Cmd == SET_CMD)
					{
					if (rdm->PDLen > DIMM_CH) rdm->PDLen= DIMM_CH;					//limit to prevent ee overflow
					memcpy(&dur.Preheat[0], &rdm->Data[0], rdm->PDLen);
					}
				else
					{
					memcpy(&rdm->Data[0], &dur.Preheat[0], DIMM_CH);
					}
				rdm->PDLen= DIMM_CH;
				respondMsg();
				break;

			case MAX_LIMIT:
				if (rdm->Cmd == SET_CMD)
					{
					if (rdm->PDLen > DIMM_CH) rdm->PDLen= DIMM_CH;					//limit to prevent ee overflow	
					memcpy(&dur.Topset[0], &rdm->Data[0], rdm->PDLen);
					}
				else
					{
					memcpy(&rdm->Data[0], &dur.Topset[0], DIMM_CH);
					}
				rdm->PDLen= DIMM_CH;
				respondMsg();
				break;

			case ERR_SCENE:
				if (rdm->Cmd == SET_CMD)
					{
					if (rdm->PDLen > DIMM_CH) rdm->PDLen= DIMM_CH;					//limit to prevent ee overflow
					memcpy(&dur.Error[0], &rdm->Data[0], rdm->PDLen);
					}
				else
					{
					memcpy(&rdm->Data[0], &dur.Error[0], DIMM_CH);
					}
				rdm->PDLen= DIMM_CH;
				respondMsg();
				break;

			default:
				rdm->Data[0]=  (uint8_t)(NR_UNKNOWN_PID>>8); //response: PID not supported!
				rdm->Data[1]=  (uint8_t)NR_UNKNOWN_PID;
				rdm->PortID= RESPONSE_TYPE_NACK;
				rdm->PDLen=  2;
				respondMsg();
				break;	
			}
		}
	RdmFlags &= ~(1<<EVAL_RDM);
	}
}



// *************** DMX Reception FSM ****************
ISR (USART_RX_vect)
{
uint8_t Temp= UCSRA;										//get state
uint8_t DmxByte= UDR;										//get data
uint8_t RxBuf= IDLE;
if (Temp &(1<<FE))											//check for break
	{
	UCSRA &= ~(1<<FE);										//reset flag
	RxBuf= BRK;												//Break detected
	}

else switch (RxState)
	{
	case BRK:
		if (DmxByte == 0x00)								//DMX start code detected
			{
			RxBuf= SB_DMX; 
			RxCount_16.u16= dur.DmxAddress;
			}	
		else if (DmxByte == 0xCC)							//RDM start code detected
			{
			RxBuf= SB_RDM; 	
			}
		break;

	case SB_DMX:
		if (--RxCount_16.u16 == 0)							//count dmx channels
			{
			RxBuf= STARTADR_DMX;							//start address reached
			DmxField[0]= DmxByte;							//get 1st ch
			RxCount_8= 1;									//seed ch counter
			}
		else
			{
			RxBuf= SB_DMX;
			}
		break;

	case STARTADR_DMX:
		DmxField[RxCount_8]= DmxByte;					//update channel
		RxCount_8++;
		if (RxCount_8 >= sizeof(DmxField))
			{
			IndFlags |= (1<<VALID_DMX);							//data set ready
			}
		else
			{
			RxBuf= STARTADR_DMX;
			}
		break;

// ********* RDM specific code *********
	case SB_RDM:
		if ((DmxByte == 0x01) && !(RdmFlags &(1<<EVAL_RDM)))
			{
			RxBuf= SSC_RDM; 				   				//valid sub start code identified and buffer free
			RxCount_8= 3;
			}
		break;

	case SSC_RDM:
		RdmField[2]= DmxByte;   				   			//get total message length
		if (DmxByte < sizeof(RdmField))
			 {
			 RxCount_16.u16= (uint16_t)(0xCD +DmxByte); 	//initialize check sum
			 RxBuf= LEN_RDM;					
			 }
		break;

	case LEN_RDM:
		RdmField[RxCount_8++]= DmxByte;						//store data
		RxCount_16.u16 += (uint16_t)DmxByte;
		if (RdmField[2] == RxCount_8)
			{
			RxBuf= CHECKH_RDM; 								//data complete?
			}
		else
			{
			RxBuf= LEN_RDM;
			}
		break;

	case CHECKH_RDM:
		if (DmxByte == RxCount_16.u8h) 						//check check sum high
			{
			RxBuf= CHECKL_RDM;
			}
		break;
	
	case CHECKL_RDM:
		if (DmxByte == RxCount_16.u8l)
			{
			RdmFlags |= (1<<EVAL_RDM);						//valid RDM packet received
			}
		break;
	}
RxState= RxBuf;
}




// *************** RDM Transmission FSM ****************

ISR (USART_TX_vect)							//send data
{
_delay_us(5);
uint8_t TxCh= gTxCh;
											//transmit RDM response
if (TxCh == 0)
	{
	UBRRL= 1;								//250kbaud
	}

else if (TxCh > TxLen)
	{
	RxState= IDLE;    						//wait for break
	UCSRB  = (1<<RXEN)|(1<<RXCIE);
	sei();									//allow interrupts in delay
	_delay_us(20);
	PORTD &= ~(1<<2);						//enable reception
						
	return;
	}

UDR= RdmField[TxCh++];						//send data
gTxCh= TxCh;
}
